@@ -55,6 +55,9 @@ typedef enum {
 /* Signal IDs */
 enum {
 	SIGNAL_ZOOM_CHANGED,
+	SIGNAL_ROTATION_CHANGED,
+	SIGNAL_NEXT_IMAGE,
+	SIGNAL_PREVIOUS_IMAGE,
 	SIGNAL_LAST
 };
 static gint view_signals [SIGNAL_LAST];
@@ -65,6 +68,20 @@ typedef enum {
 	EOG_SCROLL_VIEW_CURSOR_DRAG
 } EogScrollViewCursor;
 
+typedef enum {
+	EOG_ROTATION_0,
+	EOG_ROTATION_90,
+	EOG_ROTATION_180,
+	EOG_ROTATION_270,
+	N_EOG_ROTATIONS
+} EogRotationState;
+
+typedef enum {
+	EOG_PAN_ACTION_NONE,
+	EOG_PAN_ACTION_NEXT,
+	EOG_PAN_ACTION_PREV
+} EogPanAction;
+
 /* Drag 'n Drop */
 static GtkTargetEntry target_table[] = {
 	{ "text/uri-list", 0, 0},
@@ -160,6 +177,13 @@ struct _EogScrollViewPrivate {
 	GdkRGBA *override_bg_color;
 
 	cairo_surface_t *background_surface;
+
+	GtkGesture *pan_gesture;
+	GtkGesture *zoom_gesture;
+	GtkGesture *rotate_gesture;
+	gdouble initial_zoom;
+	EogRotationState rotate_state;
+	EogPanAction pan_action;
 };
 
 static void scroll_by (EogScrollView *view, int xofs, int yofs);
@@ -1180,8 +1204,10 @@ scroll_to (EogScrollView *view, int x, int y, gboolean change_adjustments)
 #endif
 	/* Scroll the window area and process exposure synchronously. */
 
-	gdk_window_scroll (window, -xofs, -yofs);
-	gdk_window_process_updates (window, TRUE);
+	if (!gtk_gesture_is_recognized (priv->zoom_gesture)) {
+		gdk_window_scroll (window, -xofs, -yofs);
+		gdk_window_process_updates (window, TRUE);
+	}
 
  out:
 	if (!change_adjustments)
@@ -1325,6 +1351,13 @@ set_zoom (EogScrollView *view, double zoom,
 	/* set new values */
 	priv->xofs = xofs; /* (img_width * x_rel * zoom) - anchorx; */
 	priv->yofs = yofs; /* (img_height * y_rel * zoom) - anchory; */
+
+	if (priv->dragging) {
+		priv->drag_anchor_x = anchorx;
+		priv->drag_anchor_y = anchory;
+		priv->drag_ofs_x = priv->xofs;
+		priv->drag_ofs_y = priv->yofs;
+	}
 #if 0
 	g_print ("xofs: %i  yofs: %i\n", priv->xofs, priv->yofs);
 #endif
@@ -1689,6 +1722,9 @@ eog_scroll_view_motion_event (GtkWidget *widget, GdkEventMotion *event, gpointer
 	view = EOG_SCROLL_VIEW (data);
 	priv = view->priv;
 
+	if (gtk_gesture_is_recognized (priv->zoom_gesture))
+		return TRUE;
+
 	if (!priv->dragging)
 		return FALSE;
 
@@ -1900,6 +1936,207 @@ display_draw (GtkWidget *widget, cairo_t *cr, gpointer data)
 	return TRUE;
 }
 
+static void
+zoom_gesture_begin_cb (GtkGestureZoom   *gesture,
+		       GdkEventSequence *sequence,
+		       EogScrollView    *view)
+{
+	gdouble center_x, center_y;
+	EogScrollViewPrivate *priv;
+
+	priv = view->priv;
+
+	/* Displace dragging point to gesture center */
+	gtk_gesture_get_bounding_box_center (GTK_GESTURE (gesture),
+                                             &center_x, &center_y);
+	priv->drag_anchor_x = center_x;
+	priv->drag_anchor_y = center_y;
+	priv->drag_ofs_x = priv->xofs;
+	priv->drag_ofs_y = priv->yofs;
+	priv->dragging = TRUE;
+	priv->initial_zoom = priv->zoom;
+
+        gtk_gesture_set_state (GTK_GESTURE (gesture), GTK_EVENT_SEQUENCE_CLAIMED);
+}
+
+static void
+zoom_gesture_update_cb (GtkGestureZoom   *gesture,
+			GdkEventSequence *sequence,
+			EogScrollView    *view)
+{
+	gdouble center_x, center_y, scale;
+	EogScrollViewPrivate *priv;
+
+	priv = view->priv;
+	scale = gtk_gesture_zoom_get_scale_delta (gesture);
+	gtk_gesture_get_bounding_box_center (GTK_GESTURE (gesture),
+                                             &center_x, &center_y);
+
+	drag_to (view, center_x, center_y);
+	set_zoom (view, priv->initial_zoom * scale, TRUE,
+		  center_x, center_y);
+}
+
+static void
+zoom_gesture_end_cb (GtkGestureZoom   *gesture,
+		     GdkEventSequence *sequence,
+		     EogScrollView    *view)
+{
+	EogScrollViewPrivate *priv;
+
+	priv = view->priv;
+	priv->dragging = FALSE;
+        eog_scroll_view_set_cursor (view, EOG_SCROLL_VIEW_CURSOR_NORMAL);
+}
+
+static void
+rotate_gesture_begin_cb (GtkGesture       *gesture,
+			 GdkEventSequence *sequence,
+			 EogScrollView    *view)
+{
+	EogScrollViewPrivate *priv;
+
+	priv = view->priv;
+	priv->rotate_state = EOG_TRANSFORM_NONE;
+}
+
+static void
+pan_gesture_pan_cb (GtkGesturePan   *gesture,
+		    GtkPanDirection  direction,
+		    gdouble          offset,
+		    EogScrollView   *view)
+{
+	EogScrollViewPrivate *priv;
+
+	if (eog_scroll_view_scrollbars_visible (view)) {
+		gtk_gesture_set_state (GTK_GESTURE (gesture),
+				       GTK_EVENT_SEQUENCE_DENIED);
+		return;
+	}
+
+#define PAN_ACTION_DISTANCE 200
+
+	priv = view->priv;
+	priv->pan_action = EOG_PAN_ACTION_NONE;
+	gtk_gesture_set_state (GTK_GESTURE (gesture), GTK_EVENT_SEQUENCE_CLAIMED);
+
+	if (offset > PAN_ACTION_DISTANCE) {
+		if (direction == GTK_PAN_DIRECTION_LEFT ||
+		    gtk_widget_get_direction (GTK_WIDGET (view)) == GTK_TEXT_DIR_RTL)
+			priv->pan_action = EOG_PAN_ACTION_NEXT;
+		else
+			priv->pan_action = EOG_PAN_ACTION_PREV;
+	}
+#undef PAN_ACTION_DISTANCE
+}
+
+static void
+pan_gesture_end_cb (GtkGesture       *gesture,
+		    GdkEventSequence *sequence,
+		    EogScrollView    *view)
+{
+	EogScrollViewPrivate *priv;
+
+	if (!gtk_gesture_handles_sequence (gesture, sequence))
+		return;
+
+	priv = view->priv;
+
+	if (priv->pan_action == EOG_PAN_ACTION_PREV)
+		g_signal_emit (view, view_signals [SIGNAL_PREVIOUS_IMAGE], 0);
+	else if (priv->pan_action == EOG_PAN_ACTION_NEXT)
+		g_signal_emit (view, view_signals [SIGNAL_NEXT_IMAGE], 0);
+
+	priv->pan_action = EOG_PAN_ACTION_NONE;
+}
+
+static gboolean
+scroll_view_check_angle (gdouble angle,
+			 gdouble min,
+			 gdouble max,
+			 gdouble threshold)
+{
+	if (min < max) {
+		return (angle > min - threshold &&
+			angle < max + threshold);
+	} else {
+		return (angle < max + threshold ||
+			angle > min - threshold);
+	}
+}
+
+static EogRotationState
+scroll_view_get_rotate_state (EogScrollView *view,
+			      gdouble        delta)
+{
+	EogScrollViewPrivate *priv;
+
+	priv = view->priv;
+
+#define THRESHOLD (G_PI / 16)
+	switch (priv->rotate_state) {
+	case EOG_ROTATION_0:
+		if (scroll_view_check_angle (delta, G_PI * 7 / 4,
+					     G_PI / 4, THRESHOLD))
+			return priv->rotate_state;
+		break;
+	case EOG_ROTATION_90:
+		if (scroll_view_check_angle (delta, G_PI / 4,
+					     G_PI * 3 / 4, THRESHOLD))
+			return priv->rotate_state;
+		break;
+	case EOG_ROTATION_180:
+		if (scroll_view_check_angle (delta, G_PI * 3 / 4,
+					     G_PI * 5 / 4, THRESHOLD))
+			return priv->rotate_state;
+		break;
+	case EOG_ROTATION_270:
+		if (scroll_view_check_angle (delta, G_PI * 5 / 4,
+					     G_PI * 7 / 4, THRESHOLD))
+			return priv->rotate_state;
+		break;
+	default:
+		g_assert_not_reached ();
+	}
+
+#undef THRESHOLD
+
+	if (scroll_view_check_angle (delta, G_PI / 4, G_PI * 3 / 4, 0))
+		return EOG_ROTATION_90;
+	else if (scroll_view_check_angle (delta, G_PI * 3 / 4, G_PI * 5 / 4, 0))
+		return EOG_ROTATION_180;
+	else if (scroll_view_check_angle (delta, G_PI * 5 / 4, G_PI * 7 / 4, 0))
+		return EOG_ROTATION_270;
+
+	return EOG_ROTATION_0;
+}
+
+static void
+rotate_gesture_angle_changed_cb (GtkGestureRotate *rotate,
+				 gdouble           angle,
+				 gdouble           delta,
+				 EogScrollView    *view)
+{
+	EogRotationState rotate_state;
+	EogScrollViewPrivate *priv;
+	gint angle_diffs [N_EOG_ROTATIONS][N_EOG_ROTATIONS] = {
+		{ 0,   90,  180, 270 },
+		{ 270, 0,   90,  180 },
+		{ 180, 270, 0,   90 },
+		{ 90,  180, 270, 0 }
+	};
+	gint rotate_angle;
+
+	priv = view->priv;
+	rotate_state = scroll_view_get_rotate_state (view, delta);
+
+	if (priv->rotate_state == rotate_state)
+		return;
+
+	rotate_angle = angle_diffs[priv->rotate_state][rotate_state];
+	g_signal_emit (view, view_signals [SIGNAL_ROTATION_CHANGED], 0, (gdouble) rotate_angle);
+	priv->rotate_state = rotate_state;
+}
 
 /*==================================
 
@@ -2486,6 +2723,7 @@ eog_scroll_view_init (EogScrollView *view)
 			       | GDK_BUTTON_RELEASE_MASK
 			       | GDK_POINTER_MOTION_MASK
 			       | GDK_POINTER_MOTION_HINT_MASK
+			       | GDK_TOUCH_MASK
 			       | GDK_SCROLL_MASK
 			       | GDK_KEY_PRESS_MASK);
 	g_signal_connect (G_OBJECT (priv->display), "configure_event",
@@ -2552,6 +2790,38 @@ eog_scroll_view_init (EogScrollView *view)
 			 "antialiasing-out", G_SETTINGS_BIND_GET);
 
 	g_object_unref (settings);
+
+	priv->zoom_gesture = gtk_gesture_zoom_new (GTK_WIDGET (view));
+	g_signal_connect (priv->zoom_gesture, "begin",
+			  G_CALLBACK (zoom_gesture_begin_cb), view);
+	g_signal_connect (priv->zoom_gesture, "update",
+			  G_CALLBACK (zoom_gesture_update_cb), view);
+	g_signal_connect (priv->zoom_gesture, "end",
+			  G_CALLBACK (zoom_gesture_end_cb), view);
+	g_signal_connect (priv->zoom_gesture, "cancel",
+			  G_CALLBACK (zoom_gesture_end_cb), view);
+	gtk_event_controller_set_propagation_phase (GTK_EVENT_CONTROLLER (priv->zoom_gesture),
+						    GTK_PHASE_CAPTURE);
+
+	priv->rotate_gesture = gtk_gesture_rotate_new (GTK_WIDGET (view));
+	gtk_gesture_group (priv->rotate_gesture, priv->zoom_gesture);
+	g_signal_connect (priv->rotate_gesture, "angle-changed",
+			  G_CALLBACK (rotate_gesture_angle_changed_cb), view);
+	g_signal_connect (priv->rotate_gesture, "begin",
+			  G_CALLBACK (rotate_gesture_begin_cb), view);
+	gtk_event_controller_set_propagation_phase (GTK_EVENT_CONTROLLER (priv->rotate_gesture),
+						    GTK_PHASE_CAPTURE);
+
+	priv->pan_gesture = gtk_gesture_pan_new (GTK_WIDGET (view),
+						 GTK_ORIENTATION_HORIZONTAL);
+	g_signal_connect (priv->pan_gesture, "pan",
+			  G_CALLBACK (pan_gesture_pan_cb), view);
+	g_signal_connect (priv->pan_gesture, "end",
+			  G_CALLBACK (pan_gesture_end_cb), view);
+	gtk_gesture_single_set_touch_only (GTK_GESTURE_SINGLE (priv->pan_gesture), 
+					   TRUE);
+	gtk_event_controller_set_propagation_phase (GTK_EVENT_CONTROLLER (priv->pan_gesture),
+						    GTK_PHASE_CAPTURE);
 }
 
 static void
@@ -2594,6 +2864,21 @@ eog_scroll_view_dispose (GObject *object)
 
 	free_image_resources (view);
 
+	if (priv->zoom_gesture) {
+		g_object_unref (priv->zoom_gesture);
+		priv->zoom_gesture = NULL;
+	}
+
+	if (priv->rotate_gesture) {
+		g_object_unref (priv->rotate_gesture);
+		priv->rotate_gesture = NULL;
+	}
+
+	if (priv->pan_gesture) {
+		g_object_unref (priv->pan_gesture);
+		priv->pan_gesture = NULL;
+	}
+
 	G_OBJECT_CLASS (eog_scroll_view_parent_class)->dispose (object);
 }
 
@@ -2824,6 +3109,32 @@ eog_scroll_view_class_init (EogScrollViewClass *klass)
 			      g_cclosure_marshal_VOID__DOUBLE,
 			      G_TYPE_NONE, 1,
 			      G_TYPE_DOUBLE);
+	view_signals [SIGNAL_ROTATION_CHANGED] =
+		g_signal_new ("rotation-changed",
+			      EOG_TYPE_SCROLL_VIEW,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EogScrollViewClass, rotation_changed),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__DOUBLE,
+			      G_TYPE_NONE, 1,
+			      G_TYPE_DOUBLE);
+
+	view_signals [SIGNAL_NEXT_IMAGE] =
+		g_signal_new ("next-image",
+			      EOG_TYPE_SCROLL_VIEW,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EogScrollViewClass, next_image),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__VOID,
+			      G_TYPE_NONE, 0);
+	view_signals [SIGNAL_PREVIOUS_IMAGE] =
+		g_signal_new ("previous-image",
+			      EOG_TYPE_SCROLL_VIEW,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EogScrollViewClass, previous_image),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__VOID,
+			      G_TYPE_NONE, 0);
 
 	widget_class->size_allocate = eog_scroll_view_size_allocate;
 	widget_class->style_set = eog_scroll_view_style_set;
